/*ckwg +29
 * Copyright 2016 by Kitware, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 *  * Neither name of Kitware, Inc. nor the names of any contributors may be used
 *    to endorse or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * \file
 * \brief Implementation for hough_circle_detector
 */

#include "hough_circle_detector.h"

#include <arrows/ocv/image_container.h>

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>

namespace kwiver {
namespace arrows {
namespace ocv {

// ----------------------------------------------------------------
/**
 * @brief
 *
 */
class hough_circle_detector::priv
{
public:
  // -- CONSTRUCTORS --
  priv()
    : m_dp(1)
    , m_min_dist(100)
    , m_param1(200)
    , m_param2(100)
    , m_min_radius(0)
    , m_max_radius(0)
  {}

  ~priv()
  {}

  // Hough algorithm parameters
  double m_dp;
  double m_min_dist;
  double m_param1;
  double m_param2;
  double m_min_radius;
  double m_max_radius;

}; // end class hough_circle_detector::priv


hough_circle_detector::
hough_circle_detector()
  : d( new priv )
{ }


hough_circle_detector::
hough_circle_detector( const hough_circle_detector& other)
  : d( new priv( *other.d ) )
{ }


 hough_circle_detector::
~hough_circle_detector()
{ }


// ------------------------------------------------------------------
vital::config_block_sptr
hough_circle_detector::
get_configuration() const
{
  // Get base config from base class
  vital::config_block_sptr config = vital::algorithm::get_configuration();

  config->set_value( "dp", d->m_dp,
                     "Inverse ratio of the accumulator resolution to the image resolution. "
                     "For example, if dp=1 , the accumulator has the same resolution as the input image. "
                     "If dp=2 , the accumulator has half as big width and height." );

  config->set_value( "min_dist", d->m_min_dist,
                     "Minimum distance between the centers of the detected circles. "
                     "If the parameter is too small, multiple neighbor circles may be falsely "
                     "detected in addition to a true one. If it is too large, some circles may be missed." );

  config->set_value( "param1", d->m_param1,
                     "First method-specific parameter. In case of CV_HOUGH_GRADIENT , "
                     "it is the higher threshold of the two passed to the Canny() edge detector "
                     "(the lower one is twice smaller)." );

  config->set_value( "param2", d->m_param2,
                     "Second method-specific parameter. In case of CV_HOUGH_GRADIENT , "
                     "it is the accumulator threshold for the circle centers at the detection stage. "
                     "The smaller it is, the more false circles may be detected. Circles, "
                     "corresponding to the larger accumulator values, will be returned first." );

  config->set_value( "min_radius", d->m_min_radius, "Minimum circle radius." );

  config->set_value( "max_radius", d->m_max_radius, "Maximum circle radius." );

  return config;
}


// ------------------------------------------------------------------
void
hough_circle_detector::
set_configuration(vital::config_block_sptr config)
{
  d->m_dp         = config->get_value<double>( "dp" );
  d->m_min_dist   = config->get_value<double>( "min_dist" );
  d->m_param1     = config->get_value<double>( "param1" );
  d->m_param2     = config->get_value<double>( "param2" );
  d->m_min_radius = config->get_value<double>( "min_radius" );
  d->m_max_radius = config->get_value<double>( "max_radius" );
}


// ------------------------------------------------------------------
bool
hough_circle_detector::
check_configuration(vital::config_block_sptr config) const
{
  return true;
}


// ------------------------------------------------------------------
kwiver::vital::detected_object_set_sptr
hough_circle_detector::
detect( vital::image_container_sptr image_data) const
{
  auto detected_set = std::make_shared< kwiver::vital::detected_object_set>();

  cv::Mat src = kwiver::arrows::ocv::image_container::vital_to_ocv( image_data->get_image() );
  cv::Mat src_gray;

  // Convert it to gray
  cvtColor( src, src_gray, CV_BGR2GRAY );

  // Reduce the noise so we avoid false circle detection
  cv::GaussianBlur( src_gray, src_gray, cv::Size( 9, 9 ), 2, 2 );

  cv::vector< cv::Vec3f > circles;

  // Apply the Hough Transform to find the circles
  cv::HoughCircles( src_gray, // i: source image
                    circles, // o: detected circles
                    CV_HOUGH_GRADIENT, // i: method
                    d->m_dp, // i: dp
                    d->m_min_dist, //+ src_gray.rows / 8, // i: minDist
                    d->m_param1, // i: param1 for canny edge detector
                    d->m_param2, // i: param2 for canny edge detector
                    d->m_min_radius, // i: min radius
                    d->m_max_radius ); // i: max radius

  LOG_DEBUG( m_logger, "Detected " << circles.size() << " objects." );

  // process results
  for ( size_t i = 0; i < circles.size(); ++i )
  {
    // Center point [circles[i][0], circles[i][1]]
    // Radius circles[i][2]

    // Bounding box is center +/- radius
    kwiver::vital::bounding_box_d bbox( circles[i][0] - circles[i][2], circles[i][1] - circles[i][2],
                                        circles[i][0] + circles[i][2], circles[i][1] + circles[i][2] );

    auto dot = std::make_shared< kwiver::vital::detected_object_type >();
    dot->set_score( "circle", 1.0 );

    detected_set->add( std::make_shared< kwiver::vital::detected_object >( bbox, 1.0, dot ) );
  } // end for

  return detected_set;
}

} } } // end namespace
